/* SR3C, a symbol ranking data compressor.
 *
 * This file provides a command line tool using the compressor in
 * sr3c.h and sr3c.c.
 *
 * Author: Kenneth Oksanen <cessu@iki.fi>, 2008.
 *
 * This file is distributed under the following license:

The MIT License
Copyright (c) 2008 Helsinki University of Technology
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

 */


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "sr3c.h"


static const char *argv0;


static unsigned char header[3] = { 's', 'z', '3' };

static const char *suffix = ".sz3";


static int write_header(int fd)
{
  unsigned char *bytes = header;
  size_t n_bytes = sizeof(header);
  ssize_t n;
  int e;

  do {
    n = write(fd, bytes, n_bytes);
    if (n < 0) {
      e = errno;
      fprintf(stderr, "%s: ", argv0);
      errno = e;
      perror("Write failed");
      return n;
    }
    bytes += n;
    n_bytes -= n;
  } while (n_bytes > 0);
  
  return 0;
}


static int read_header(int fd)
{
  unsigned char buf[sizeof(header)], *bytes = buf;
  size_t n_bytes = sizeof(header);
  ssize_t n;

  do {
    n = read(fd, bytes, n_bytes);
    if (n < 0) {
      int e = errno;
      fprintf(stderr, "%s: ", argv0);
      errno = e;
      perror("Read failed");
      return n;
    }
    bytes += n;
    n_bytes -= n;
  } while (n_bytes > 0);

  if (memcmp(header, buf, sizeof(header)) != 0) {
    fprintf(stderr,
	    "%s: Attempted to compress something not compressed by sr3c.\n", 
	    argv0);
    return 1;
  }

  return 0;
}


typedef struct {
  int fd;
} cb_ctx_t;


static int write_f(const unsigned char *bytes, size_t n_bytes,
		   int flush, void *ctx)
{
  cb_ctx_t *my_ctx = (cb_ctx_t *) ctx;
  ssize_t n;

  do {
    n = write(my_ctx->fd, bytes, n_bytes);
    if (n < 0) {
      int e = errno;
      fprintf(stderr, "%s: ", argv0);
      errno = e;
      perror("Write failed");
      return n;
    }
    bytes += n;
    n_bytes -= n;
  } while (n_bytes > 0);

  if (flush
      && fdatasync(my_ctx->fd) < 0
      /* If we fdatasync on stdout we may get EINVAL, ignore it. */
      && errno != EINVAL) {
    int e = errno;
    fprintf(stderr, "%s: ", argv0);
    errno = e;
    perror("Fdatasync failed");
    return -1;
  }

  return 0;
}


static unsigned char buf[65536];


static int compress(int in_fd, sr3c_context_t *ctx, int msecs)
{
  struct timeval tv;
  ssize_t n, cum_n;
  fd_set fdset;
  int err;

  if (msecs >= 0) {
    /* Real-time operation requested, set input file to nonblocking. */
    fcntl(in_fd, F_SETFL, O_NONBLOCK);

    while (1) {
      cum_n = 0;
      while ((n = read(in_fd, buf, sizeof(buf))) > 0) {
	err = sr3c_compress(buf, n, ctx);
	if (err)
	  return err;
	cum_n += n;
      }

      if (n == 0)		/* EOF */
	return sr3c_flush(ctx);
      if (errno != EAGAIN) {	/* Some other error. */
	int e = errno;
	fprintf(stderr, "%s: ", argv0);
	errno = e;
	perror("Non-blocking read failed");
	sr3c_flush(ctx);
	return -1;
      }
      /* If we compressed too little data, flush. Theoretically the
	 maximum lag from compression to uncompression is around 20
	 bytes of intermediate uncompressed data, but we've doubled it
	 here. */
      if (cum_n < 40) {
	err = sr3c_flush(ctx);
	if (err)
	  return err;
      }
      FD_CLR(in_fd, &fdset);
      tv.tv_sec = 0;
      tv.tv_usec = 1000 * msecs;
      select(1, &fdset, NULL, NULL, &tv);
    }
    /*NOTREACHED*/
  } else {
    while ((n = read(in_fd, buf, sizeof(buf))) > 0) {
      err = sr3c_compress(buf, n, ctx);
      if (err)
	return err;
    }
    if (n < 0) {
      int e = errno;
      fprintf(stderr, "%s: ", argv0);
      errno = e;
      perror("Read failed");
    }
    return sr3c_flush(ctx);
  }
}


static int uncompress(int in_fd, sr3c_context_t *ctx)
{
  ssize_t n;
  int err;

  while ((n = read(in_fd, buf, sizeof(buf))) > 0) {
    err = sr3c_uncompress(buf, n, ctx);
    if (err)
      return err;
  }
  if (n < 0) {
    int e = errno;
    fprintf(stderr, "%s: ", argv0);
    errno = e;
    perror("Read failed");
  }
  return 0;
}


/* A convenient buffer capable for holding at least any file name and
   some message text. */
static char tmps[FILENAME_MAX + 128];


int main(int argc, char *argv[])
{
  int opt, do_uncompress = 0, to_stdout = 0, have_header = 1;
  int in_fd, out_fd, i, err;
  long msecs = -1;
  sr3c_context_t *ctx;
  cb_ctx_t cb_ctx;
  char *endp;

  argv0 = argv[0];

  while ((i = getopt(argc, argv, "cdnho:")) != -1)
    switch (i) {
    case 'c':
      to_stdout = 1;
      break;
    case 'd':
      do_uncompress = 1;
      break;
    case 'n':
      have_header = 0;
      break;
    case 'o':
      msecs = strtol(optarg, &endp, 10);
      if (endp == optarg || *endp != '\0' || msecs < 0)
	goto cmdline_error;
      break;
    case 'h':
      printf("Usage: %s [OPTIONS] [FILES]\n\
Compress or uncompress FILES in-place or in absense of FILES from stdin\n\
to stdout.  The compression algorithm is based on symbol ranking and\n\
arithmetic compression.  Compression is slightly faster than gzip -7,\n\
but the compression ratio is roughly 11%% better.\n\
\n\
Options:\n\
  -c         Write to standard output, keep original files unchanged\n\
  -d         Decompress\n\
  -n         No header, make compatible with sr3c_compress/uncompress.\n\
  -o timeout Compress/uncompress on-line.  Block for no longer than given ms.\n\
  -h         Display this help.\n\
\n\
sr3c was written by Kenneth Oksanen <cessu@iki.fi> partly based on the ideas\n\
of Peter Fenwick and Matt Mahoney.  sr3c is copyright (c) by 2008 Helsinki\n\
University of Technology and distributed under the MIT License.\n\
", argv[0]);
      return 0;
      break;
    default:
    cmdline_error:
      fprintf(stderr, "Usage: %s [-cdh] [-o timeout] [name ...]\n", argv[0]);
      exit(EXIT_FAILURE);
    }

  if (msecs >= 0 && do_uncompress)
    fprintf(stderr,
	    "Note: %s flag -o has no effect in the presence of flag -d.\n",
	    argv[0]);
  

  ctx = sr3c_alloc(write_f, &cb_ctx);
  if (ctx == NULL) {
    fprintf(stderr, "%s: sr3c_alloc failed, maybe out of memory.\n", argv0);
    exit(EXIT_FAILURE);
  }
  
  if (optind == argc) {
    /* Compress from stdin to stdout. */
    cb_ctx.fd = 1;
    if (do_uncompress) {
      if (have_header) {
	err = read_header(0);
	if (err)
	  goto out;
      }
      err = uncompress(0, ctx);
    } else {
      if (have_header) {
	err = write_header(1);
	if (err)
	  goto out;
      }
      err = compress(0, ctx, msecs);
    }
  } else if (to_stdout) {
    /* Catenate all input files and compress/uncompress. */
    cb_ctx.fd = 1;
    err = 0;
    for (i = optind; !err && i < argc; i++) {
      in_fd = open(argv[i], O_RDONLY, 0);
      if (in_fd < 0) {
	int e = errno;
	fprintf(stderr, "%s: ", argv0);
	snprintf(tmps, sizeof(tmps), "Failed to open file '%s' for reading", 
		 argv[i]);
	errno = e;
	perror(tmps);
	err = 1;
	goto out;
      }
      if (do_uncompress) {
	if (have_header) {
	  err = read_header(0);
	  if (!err)
	    err = uncompress(0, ctx);
	} else
	  err = uncompress(0, ctx);
      } else {
	if (have_header) {
	  err = write_header(1);
	  if (!err)
	    err = compress(0, ctx, msecs);
	} else
	  err = compress(0, ctx, msecs);
      }
      close(in_fd);
    }
  } else if (do_uncompress) {
    /* Uncompress given files in place, removing the suffix .sz3
       and deleting originals when successful. */
    for (i = optind; i < argc; i++) {
      int tmpslen = strlen(argv[i]) - strlen(suffix);
      if (tmpslen >= sizeof(tmps)) {
	fprintf(stderr, "%s: too long filename `%s', skipping.\n", 
		argv0, argv[i]);
	continue;
      }
      if (strcmp(argv[i] + tmpslen, suffix) != 0) {
	fprintf(stderr, "%s: missing suffix `%s' in `%s', skipping.\n", 
		argv0, suffix, argv[i]);
	continue;
      }
      memcpy(tmps, argv[i], tmpslen);
      tmps[tmpslen] = '\0';
      in_fd = open(argv[i], O_RDONLY, 0);
      if (in_fd < 0) {
	int e = errno;
	fprintf(stderr, "%s: ", argv0);
	snprintf(tmps, sizeof(tmps), "Failed to open file '%s' for reading", 
		 argv[i]);
	errno = e;
	perror(tmps);
	err = 1;
	continue;
      }
      cb_ctx.fd = creat(tmps, S_IRUSR | S_IWUSR);
      if (cb_ctx.fd < 0) {
	int e = errno;
	fprintf(stderr, "%s: ", argv0);
	snprintf(tmps, sizeof(tmps), "Failed to open file '%s%s' for writing", 
		 argv[i], suffix);
	errno = e;
	perror(tmps);
	err = 1;
	close(in_fd);
	continue;
      }
      sr3c_reset(ctx, NULL, NULL);
      if (have_header) {
	err = read_header(in_fd);
	if (!err)
	  err = uncompress(in_fd, ctx);
      } else
	err = uncompress(in_fd, ctx);
      close(in_fd);
      if (err) {
	/* Something went wrong. Close and remove the compressed file. */
	close(cb_ctx.fd);
	unlink(tmps);
      } else {
	/* So far so good. Close the compressed file, unlink source file. */
	err = close(cb_ctx.fd);
	if (err) {
	  int e = errno;
	  fprintf(stderr, "%s: ", argv0);
	  errno = e;
	  perror("Close failed");
	  unlink(tmps);
	  continue;
	}
	unlink(argv[i]);
      }
    }
  } else {
    /* Compress given files in place, adding the suffix .sz3
       and deleting originals when successful. */
    for (i = optind; i < argc; i++) {
      if (strlen(argv[i]) + strlen(suffix) + 1 > sizeof(tmps)) {
	fprintf(stderr, "%s: too long filename `%s', skipping.\n", 
		argv0, argv[i]);
	continue;
      }
      if (strlen(argv[i]) > strlen(suffix)
	  && strcmp(argv[i] + strlen(argv[i]) - strlen(suffix), suffix) == 0) {
	fprintf(stderr, "%s: `%s' already seems compressed, skipping.\n", 
		argv0, argv[i]);
	continue;
      }
      in_fd = open(argv[i], O_RDONLY, 0);
      if (in_fd < 0) {
	int e = errno;
	fprintf(stderr, "%s: ", argv0);
	snprintf(tmps, sizeof(tmps), "Failed to open file '%s' for reading", 
		 argv[i]);
	errno = e;
	perror(tmps);
	err = 1;
	continue;
      }
      sprintf(tmps, "%s%s", argv[i], suffix);
      cb_ctx.fd = creat(tmps, S_IRUSR | S_IWUSR);
      if (cb_ctx.fd < 0) {
	int e = errno;
	fprintf(stderr, "%s: ", argv0);
	snprintf(tmps, sizeof(tmps), "Failed to open file '%s%s' for writing", 
		 argv[i], suffix);
	errno = e;
	perror(tmps);
	err = 1;
	close(in_fd);
	continue;
      }
      sr3c_reset(ctx, NULL, NULL);
      if (have_header) {
	err = write_header(cb_ctx.fd);
	if (!err)
	  err = compress(in_fd, ctx, msecs);
      } else
	err = compress(in_fd, ctx, msecs);
      close(in_fd);
      if (err) {
	/* Something went wrong. Close and remove the compressed file. */
	close(cb_ctx.fd);
	unlink(tmps);
      } else {
	/* So far so good. Close the compressed file, unlink source file. */
	err = close(cb_ctx.fd);
	if (err) {
	  int e = errno;
	  fprintf(stderr, "%s: ", argv0);
	  errno = e;
	  perror("Close failed");
	  unlink(tmps);
	  continue;
	}
	unlink(argv[i]);
      }
    }
  }

 out:
  sr3c_free(ctx);

  return err == 0 ? 0 : EXIT_FAILURE;
}
